from lake.models.model import Model


# transpose buffer model
class TBModel(Model):

    def __init__(self,
                 word_width,
                 fetch_width,
                 num_tb,
                 tb_height,
                 max_range,
                 max_range_inner):

        # generation parameters
        self.word_width = word_width
        self.fetch_width = fetch_width
        self.num_tb = num_tb
        self.tb_height = tb_height
        self.max_range = max_range
        self.max_range_inner = max_range_inner

        # configuration registers
        self.config = {}
        self.config["range_outer"] = 1
        self.config["range_inner"] = 1
        self.config["stride"] = 1
        self.config["indices"] = [0]
        self.config["dimensionality"] = 1
        self.config["starting_addr"] = 0

        # initialize transpose buffer
        self.tb = []
        if self.fetch_width == 1:
            for i in range(2 * self.tb_height):
                self.tb.append(0)
        else:
            for i in range(2 * self.tb_height):
                row = []
                for j in range(self.fetch_width):
                    row.append(0)
                self.tb.append(row)

        self.tb_valid = []
        for i in range(2 * self.tb_height):
            self.tb_valid.append(0)

        self.mask_valid = 0

        self.row_index = 0
        self.input_buf_index = 0
        self.out_buf_index = 1
        self.prev_out_buf_index = 0

        self.col_pixels = []
        for i in range(self.tb_height):
            self.col_pixels.append(0)

        self.output_index = 0
        self.index_inner = 0
        self.index_outer = 0
        self.curr_out_start = 0
        self.output_valid = 0
        self.pause_tb = 1
        self.pause_output = 0
        self.rdy_to_arbiter = 1
        self.start_data = 0
        self.old_start_data = 0
        self.output_index_abs = 0
        self.prev_out2 = 1
        self.prev_col_pixels2 = []
        self.prev_col_pixels3 = []
        self.on_next_line = 0
        self.switch_next_line = 0
        self.switch_out_buf = 0

    def set_config(self, new_config):
        for key, config_val in new_config.items():
            if key not in self.config:
                AssertionError("Gave bad config...")
            else:
                self.config[key] = config_val

        self.tb = []
        if self.fetch_width == 1:
            for i in range(2 * self.tb_height):
                self.tb.append(0)
        else:
            for i in range(2 * self.tb_height):
                row = []
                for j in range(self.fetch_width):
                    row.append(0)
                self.tb.append(row)

        self.tb_valid = []
        for i in range(2 * self.tb_height):
            self.tb_valid.append(0)

        self.mask_valid = 0

        self.row_index = 0
        self.input_buf_index = 0
        self.out_buf_index = 1
        self.prev_out_buf_index = 0

        self.col_pixels = []
        for i in range(self.tb_height):
            self.col_pixels.append(0)

        self.output_index = 0
        self.index_inner = 0
        self.index_outer = 0
        self.curr_out_start = 0
        self.output_valid = 0
        self.pause_tb = 1
        self.pause_output = 0
        self.rdy_to_arbiter = 1
        self.prev_col_pixels = 0
        self.start_data = 0
        self.old_start_data = 0
        self.output_index_abs = 0
        self.on_next_line = 0
        self.switch_next_line = 0
        self.switch_out_buf = 0

    def get_col_pixels(self):
        return self.col_pixels

    def get_output_valid(self):
        return self.output_valid

    def get_rdy_to_arbiter(self):
        return self.rdy_to_arbiter

    def output_from_tb(self, input_data, valid_data, ack_in, ren, mem_valid_data):

        # Grab current values...
        index_inner_curr = self.index_inner
        index_outer_curr = self.index_outer
        pause_tb_curr = self.pause_tb
        row_index_curr = self.row_index
        input_buf_index_curr = self.input_buf_index
        # tb
        # col pixels
        output_index_curr = self.output_index
        output_valid_curr = self.output_valid
        out_buf_index_curr = self.out_buf_index
        curr_out_start_curr = self.curr_out_start
        prev_out_buf_index_curr = self.prev_out_buf_index
        start_data_curr = self.start_data
        old_start_data_curr = self.old_start_data
        rdy_to_arbiter_curr = self.rdy_to_arbiter
        on_next_line_curr = self.on_next_line
        tb_curr = self.tb.copy()

        # Combinational updates

        if ((index_outer_curr == 0) and (not on_next_line_curr) and
                ((self.config["dimensionality"] == 1) or
                    ((self.config["dimensionality"] == 2) & (index_inner_curr == 0)))):
            self.switch_next_line = 1
        else:
            self.switch_next_line = 0

        if pause_tb_curr:
            self.pause_output = 1
        else:
            self.pause_output = 1 - ren

        if self.config["dimensionality"] == 1:
            self.output_index_abs = index_outer_curr * self.config["stride"] + self.config["starting_addr"]
        else:
            self.output_index_abs = index_outer_curr * self.config["stride"] + \
                self.config["indices"][index_inner_curr] + \
                self.config["starting_addr"]

        self.output_index = self.output_index_abs % self.fetch_width

        if self.switch_next_line | (self.output_index_abs >= curr_out_start_curr + self.fetch_width):
            self.switch_out_buf = 1
        else:
            self.switch_out_buf = 0

        # output from tb
        if self.config["dimensionality"] == 0:
            self.col_pixels = [0]
        else:
            self.col_pixels = []
            for i in range(self.tb_height):
                if self.fetch_width == 1:
                    self.col_pixels.append(tb_curr[i + self.tb_height * (1 -
                                                   (out_buf_index_curr ^
                                                       self.switch_out_buf))])
                else:
                    self.col_pixels.append(
                        tb_curr[i + self.tb_height * (1 - (out_buf_index_curr ^
                                self.switch_out_buf))][self.output_index])

        # Set mask valid
        if out_buf_index_curr ^ self.switch_out_buf:
            self.mask_valid = self.tb_valid[0]
        else:
            self.mask_valid = self.tb_valid[1]

        # Set output valid
        if self.config["dimensionality"] == 0:
            self.output_valid = 0
        elif self.pause_output:
            self.output_valid = 0
        else:
            self.output_valid = self.mask_valid

        # Stateful updates...

        if self.config["dimensionality"] != 0:
            # Index outer
            if self.config["dimensionality"] == 1:
                if index_outer_curr == self.config["range_outer"] - 1:
                    if not self.pause_output:
                        self.index_outer = 0
                    else:
                        self.index_outer = index_outer_curr
                elif pause_tb_curr:
                    self.index_outer = index_outer_curr
                elif not self.pause_output:
                    self.index_outer = index_outer_curr + 1
            else:
                if index_inner_curr == self.config["range_inner"] - 1:
                    if index_outer_curr == self.config["range_outer"] - 1:
                        if not self.pause_output:
                            self.index_outer = 0
                        else:
                            self.index_outer = index_outer_curr
                    elif not self.pause_output:
                        self.index_outer = index_outer_curr + 1
            # Index inner
            if self.config["dimensionality"] == 1:
                self.index_inner = 0
            else:
                if index_inner_curr == self.config["range_inner"] - 1:
                    if not self.pause_output:
                        self.index_inner = 0
                elif pause_tb_curr:
                    self.index_inner = index_inner_curr
                elif not self.pause_output:
                    self.index_inner = index_inner_curr + 1

            # Pause tb
            if index_outer_curr == self.config["range_outer"] - 1:
                if (self.config["dimensionality"] == 1) or \
                        ((self.config["dimensionality"] == 2) and
                            (index_inner_curr == self.config["range_inner"] - 1)):
                    if not self.pause_output:
                        if (not rdy_to_arbiter_curr) or valid_data:
                            self.pause_tb = 0
                        else:
                            self.pause_tb = 1
            elif pause_tb_curr:
                self.pause_tb = 1 - valid_data

            # Row index + input_buf_index + input tb
            if valid_data:
                self.tb[row_index_curr + self.tb_height * input_buf_index_curr] = input_data
                self.tb_valid[row_index_curr + self.tb_height * input_buf_index_curr] = mem_valid_data
                if row_index_curr == self.tb_height - 1:
                    self.row_index = 0
                    self.input_buf_index = 1 - input_buf_index_curr
                else:
                    self.row_index = row_index_curr + 1

            # Set out buf index
            if not start_data_curr:
                self.out_buf_index = 1
            elif self.config["dimensionality"] == 1:
                if (index_outer_curr == 0) and (not on_next_line_curr):
                    self.out_buf_index = 1 - out_buf_index_curr
                elif (self.output_index_abs >= curr_out_start_curr + self.fetch_width):
                    self.out_buf_index = 1 - out_buf_index_curr
            elif self.config["dimensionality"] == 2:
                if (index_inner_curr == 0) and (index_outer_curr == 0) and (not on_next_line_curr):
                    self.out_buf_index = 1 - out_buf_index_curr
                elif (self.output_index_abs >= curr_out_start_curr + self.fetch_width):
                    self.out_buf_index = 1 - out_buf_index_curr

            # Set on next line
            if self.config["dimensionality"] == 1:
                if (index_outer_curr == 0) and (not on_next_line_curr):
                    self.on_next_line = 1
                elif (index_outer_curr == self.config["range_outer"] - 1):
                    self.on_next_line = 0
            elif self.config["dimensionality"] == 2:
                if (index_inner_curr == 0) and (index_outer_curr == 0) and (not on_next_line_curr):
                    self.on_next_line = 1
                elif (index_inner_curr == self.config["range_inner"] - 1):
                    if (index_outer_curr == self.config["range_outer"] - 1):
                        self.on_next_line = 0

            # Set curr out start
            if index_outer_curr == 0:
                if self.config["dimensionality"] == 1:
                    self.curr_out_start = 0
                elif (self.config["dimensionality"] == 2) and (index_inner_curr == 0):
                    self.curr_out_start = 0
            elif self.output_index_abs >= curr_out_start_curr + self.fetch_width:
                self.curr_out_start = curr_out_start_curr + self.fetch_width

            # Set pref out buf index
            self.prev_out_buf_index = out_buf_index_curr

            # Set start data
            if valid_data and (not start_data_curr):
                self.start_data = 1

            # Set old start data
            self.old_start_data = start_data_curr

            # Set rdy to arbiter
            if start_data_curr and not old_start_data_curr:
                self.rdy_to_arbiter = 1
            elif self.switch_out_buf:
                self.rdy_to_arbiter = 1
            elif self.tb_height != 1:
                if row_index_curr != self.tb_height - 1:
                    self.rdy_to_arbiter = 1
            elif ack_in:
                self.rdy_to_arbiter = 0

    def print_tb(self, input_data, valid_data, ack_in, ren, mem_valid_data):
        print("INPUTS")

        print("input data: ", input_data)
        print("valid data: ", valid_data)
        print("ack in ", ack_in)
        print("ren ", ren)
        print("mem valid data", mem_valid_data)

        print("input index ", self.tb_height * self.input_buf_index + self.row_index)
        print("tb: ", self.tb)
        print("row index: ", self.row_index)
        print("input buf index: ", self.input_buf_index)
        print("out buf index ", self.out_buf_index)
        print("prev out buf index ", self.prev_out_buf_index)
        print("col pixels ", self.col_pixels)
        print("prev col pixels ", self.prev_col_pixels)
        print("output index abs", self.output_index_abs)
        print("output index ", self.output_index)
        print("index inner ", self.index_inner)
        print("index outer ", self.index_outer)
        print("curr out start ", self.curr_out_start)
        print("output valid ", self.output_valid)
        print("pause tb ", self.pause_tb)
        print("pause output ", self.pause_output)
        print("rdy ", self.rdy_to_arbiter)
        print("start data", self.start_data)
        print("old start data", self.old_start_data)
        print("col pixels", self.col_pixels)
        print("out valid", self.output_valid)
        print("switch out buf", self.switch_out_buf)
        print("mask valid", self.mask_valid)
        print("tb valid", self.tb_valid)

    def interact(self, input_data, valid_data, ack_in, ren, mem_valid_data):
        # print("before")
        # self.print_tb(input_data, valid_data, ack_in, ren)
        self.output_from_tb(input_data, valid_data, ack_in, ren, mem_valid_data)
        # print(" ")
        # print("after")
        # self.print_tb(input_data, valid_data, ack_in, ren, mem_valid_data)
        # print(" ")
        return self.col_pixels, self.output_valid, self.rdy_to_arbiter



from kratos import *
from lake.attributes.config_reg_attr import ConfigRegAttr
from lake.passes.passes import lift_config_reg
from lake.modules.reg_fifo import RegFIFO
import kratos as kts


class StrgFIFO(Generator):
    '''
    Storage Fifo
        This module performs FIFO using multiple banks of storage and
        some small register fifos in the core

        Note: for this module to function, we assume that there is an address
        generator on both the output and input that uses a bank interleaving pattern.
        In this way, we can prevent any conflicts from occuring
    '''
    def __init__(self,
                 data_width=16,
                 banks=1,
                 memory_width=64,
                 rw_same_cycle=False,
                 read_delay=1,
                 addr_width=9):
        super().__init__("strg_fifo")

        # Generation parameters
        self.banks = banks
        self.data_width = data_width
        self.memory_width = memory_width
        self.rw_same_cycle = rw_same_cycle
        self.read_delay = read_delay
        self.addr_width = addr_width
        self.fw_int = int(self.memory_width / self.data_width)

        # assert banks > 1 or rw_same_cycle is True or self.fw_int > 1, \
        #     "Can't sustain throughput with this setup. Need potential bandwidth for " + \
        #     "1 write and 1 read in a cycle - try using more banks or a macro that supports 1R1W"

        # Clock and Reset
        self._clk = self.clock("clk")
        self._rst_n = self.reset("rst_n")

        # Inputs + Outputs
        self._push = self.input("push", 1)
        self._data_in = self.input("data_in", self.data_width)
        self._pop = self.input("pop", 1)

        self._data_out = self.output("data_out", self.data_width)
        self._valid_out = self.output("valid_out", 1)
        self._empty = self.output("empty", 1)
        self._full = self.output("full", 1)

        # get relevant signals from the storage banks
        self._data_from_strg = self.input("data_from_strg", self.data_width,
                                          size=(self.banks,
                                                self.fw_int),
                                          explicit_array=True,
                                          packed=True)

        self._wen_addr = self.var("wen_addr", self.addr_width,
                                  size=self.banks,
                                  explicit_array=True,
                                  packed=True)

        self._ren_addr = self.var("ren_addr", self.addr_width,
                                  size=self.banks,
                                  explicit_array=True,
                                  packed=True)

        self._front_combined = self.var("front_combined", self.data_width,
                                        size=self.fw_int,
                                        explicit_array=True,
                                        packed=True)

        self._data_to_strg = self.output("data_to_strg", self.data_width,
                                         size=(self.banks,
                                               self.fw_int),
                                         explicit_array=True,
                                         packed=True)

        self._wen_to_strg = self.output("wen_to_strg", self.banks)
        self._ren_to_strg = self.output("ren_to_strg", self.banks)

        self._num_words_mem = self.var("num_words_mem", self.data_width)

        if self.banks == 1:
            self._curr_bank_wr = self.var("curr_bank_wr", 1)
            self.wire(self._curr_bank_wr, kts.const(0, 1))
            self._curr_bank_rd = self.var("curr_bank_rd", 1)
            self.wire(self._curr_bank_rd, kts.const(0, 1))
        else:
            self._curr_bank_wr = self.var("curr_bank_wr", kts.clog2(self.banks))
            self._curr_bank_rd = self.var("curr_bank_rd", kts.clog2(self.banks))

        self._write_queue = self.var("write_queue", self.data_width,
                                     size=(self.banks,
                                           self.fw_int),
                                     explicit_array=True,
                                     packed=True)

        # Lets us know if the bank has a write queued up
        self._queued_write = self.var("queued_write", self.banks)

        self._front_data_out = self.var("front_data_out", self.data_width)
        self._front_pop = self.var("front_pop", 1)
        self._front_empty = self.var("front_empty", 1)
        self._front_full = self.var("front_full", 1)
        self._front_valid = self.var("front_valid", 1)
        self._front_par_read = self.var("front_par_read", 1)
        self._front_par_out = self.var("front_par_out", self.data_width,
                                       size=(self.fw_int,
                                             1),
                                       explicit_array=True,
                                       packed=True)

        self._front_rd_ptr = self.var("front_rd_ptr", max(1, clog2(self.fw_int)))

        self._front_push = self.var("front_push", 1)
        self.wire(self._front_push, self._push & (~self._full | self._pop))

        self._front_rf = RegFIFO(data_width=self.data_width,
                                 width_mult=1,
                                 depth=self.fw_int,
                                 parallel=True,
                                 break_out_rd_ptr=True)

        # This one breaks out the read pointer so we can properly
        # reorder the data to storage

        self.add_child("front_rf", self._front_rf,
                       clk=self._clk,
                       clk_en=kts.const(1, 1),
                       rst_n=self._rst_n,
                       push=self._front_push,
                       pop=self._front_pop,
                       empty=self._front_empty,
                       full=self._front_full,
                       valid=self._front_valid,
                       parallel_read=self._front_par_read,
                       parallel_load=kts.const(0, 1),  # We don't need to parallel load the front
                       parallel_in=0,  # Same reason as above
                       parallel_out=self._front_par_out,
                       num_load=0,
                       rd_ptr_out=self._front_rd_ptr)
        self.wire(self._front_rf.ports.data_in[0], self._data_in)
        self.wire(self._front_data_out, self._front_rf.ports.data_out[0])

        self._back_data_in = self.var("back_data_in", self.data_width)
        self._back_data_out = self.var("back_data_out", self.data_width)
        self._back_push = self.var("back_push", 1)
        self._back_empty = self.var("back_empty", 1)
        self._back_full = self.var("back_full", 1)
        self._back_valid = self.var("back_valid", 1)
        self._back_pl = self.var("back_pl", 1)
        self._back_par_in = self.var("back_par_in", self.data_width,
                                     size=(self.fw_int,
                                           1),
                                     explicit_array=True,
                                     packed=True)
        self._back_num_load = self.var("back_num_load", clog2(self.fw_int) + 1)

        self._back_occ = self.var("back_occ", clog2(self.fw_int) + 1)
        self._front_occ = self.var("front_occ", clog2(self.fw_int) + 1)

        self._back_rf = RegFIFO(data_width=self.data_width,
                                width_mult=1,
                                depth=self.fw_int,
                                parallel=True,
                                break_out_rd_ptr=False)

        self._fw_is_1 = self.var("fw_is_1", 1)
        self.wire(self._fw_is_1, kts.const(self.fw_int == 1, 1))

        self._back_pop = self.var("back_pop", 1)
        if self.fw_int == 1:
            self.wire(self._back_pop, self._pop & (~self._empty | self._push) & ~self._back_pl)
        else:
            self.wire(self._back_pop, self._pop & (~self._empty | self._push))

        self.add_child("back_rf", self._back_rf,
                       clk=self._clk,
                       clk_en=kts.const(1, 1),
                       rst_n=self._rst_n,
                       push=self._back_push,
                       pop=self._back_pop,
                       empty=self._back_empty,
                       full=self._back_full,
                       valid=self._back_valid,
                       parallel_read=kts.const(0, 1),
                       # Only do back load when data is going there
                       parallel_load=self._back_pl & self._back_num_load.r_or(),
                       parallel_in=self._back_par_in,
                       num_load=self._back_num_load)
        self.wire(self._back_rf.ports.data_in[0], self._back_data_in)
        self.wire(self._back_data_out, self._back_rf.ports.data_out[0])
        # send the writes through when a read isn't happening
        for i in range(self.banks):
            self.add_code(self.send_writes, idx=i)
            self.add_code(self.send_reads, idx=i)

        # Set the parallel load to back bank - if no delay it's immediate
        # if not, it's delayed :)
        if self.read_delay == 1:
            self._ren_delay = self.var("ren_delay", 1)
            self.add_code(self.set_parallel_ld_delay_1)
            self.wire(self._back_pl, self._ren_delay)
        else:
            self.wire(self._back_pl, self._ren_to_strg.r_or())

        # Combine front end data - just the items + incoming
        # this data is actually based on the rd_ptr from the front fifo
        for i in range(self.fw_int):
            self.wire(self._front_combined[i], self._front_par_out[self._front_rd_ptr + i])
        # This is always true
        # self.wire(self._front_combined[self.fw_int - 1], self._data_in)

        # prioritize queued writes, otherwise send combined data
        for i in range(self.banks):
            self.wire(self._data_to_strg[i],
                      kts.ternary(self._queued_write[i], self._write_queue[i], self._front_combined))

        # Wire the thin output from front to thin input to back
        self.wire(self._back_data_in, self._front_data_out)
        self.wire(self._back_push, self._front_valid)
        self.add_code(self.set_front_pop)

        # Queue writes
        for i in range(self.banks):
            self.add_code(self.set_write_queue, idx=i)

        # Track number of words in memory
        # if self.read_delay == 1:
        #     self.add_code(self.set_num_words_mem_delay)
        # else:
        self.add_code(self.set_num_words_mem)

        # Track occupancy of the two small fifos
        self.add_code(self.set_front_occ)
        self.add_code(self.set_back_occ)

        if self.banks > 1:
            self.add_code(self.set_curr_bank_wr)
            self.add_code(self.set_curr_bank_rd)
        if self.read_delay == 1:
            self._prev_bank_rd = self.var("prev_bank_rd", max(1, kts.clog2(self.banks)))
            self.add_code(self.set_prev_bank_rd)

        # Parallel load data to back - based on num_load
        index_into = self._curr_bank_rd
        if self.read_delay == 1:
            index_into = self._prev_bank_rd
        for i in range(self.fw_int - 1):
            # Shift data over if you bypassed from the memory output
            self.wire(self._back_par_in[i],
                      kts.ternary(self._back_num_load == self.fw_int,
                                  self._data_from_strg[index_into][i],
                                  self._data_from_strg[index_into][i + 1]))
        self.wire(self._back_par_in[self.fw_int - 1],
                  kts.ternary(self._back_num_load == self.fw_int,
                              self._data_from_strg[index_into][self.fw_int - 1],
                              kts.const(0, self.data_width)))

        # Set the parallel read to the front fifo - analogous with trying to write to the memory
        self.add_code(self.set_front_par_read)

        # Set the number being parallely loaded into the register
        self.add_code(self.set_back_num_load)

        # Data out and valid out are (in the general case) just the data and valid from the back fifo
        # In the case where we have a fresh memory read, it would be from that
        bank_idx_read = self._curr_bank_rd
        if self.read_delay == 1:
            bank_idx_read = self._prev_bank_rd
        self.wire(self._data_out,
                  kts.ternary(self._back_pl, self._data_from_strg[bank_idx_read][0], self._back_data_out))
        self.wire(self._valid_out, kts.ternary(self._back_pl, self._pop, self._back_valid))

        # Set addresses to storage
        for i in range(self.banks):
            self.add_code(self.set_wen_addr, idx=i)
            self.add_code(self.set_ren_addr, idx=i)
        # Now deal with a shared address vs separate addresses
        if self.rw_same_cycle:
            # Separate
            self._wen_addr_out = self.output("wen_addr_out", self.addr_width,
                                             size=self.banks,
                                             explicit_array=True,
                                             packed=True)
            self._ren_addr_out = self.output("ren_addr_out", self.addr_width,
                                             size=self.banks,
                                             explicit_array=True,
                                             packed=True)
            self.wire(self._wen_addr_out, self._wen_addr)
            self.wire(self._ren_addr_out, self._ren_addr)
        else:
            self._addr_out = self.output("addr_out", self.addr_width,
                                         size=self.banks,
                                         explicit_array=True,
                                         packed=True)
            # If sharing the addresses, send read addr with priority
            for i in range(self.banks):
                self.wire(self._addr_out[i],
                          kts.ternary(self._wen_to_strg[i], self._wen_addr[i], self._ren_addr[i]))

        # Do final empty/full
        self._num_items = self.var("num_items", self.data_width)
        self.add_code(self.set_num_items)
        self._fifo_depth = self.input("fifo_depth", self.data_width)
        self._fifo_depth.add_attribute(ConfigRegAttr("Fifo depth..."))
        self.wire(self._empty, self._num_items == 0)
        self.wire(self._full, self._num_items == (self._fifo_depth))

    @always_comb
    def send_writes(self, idx):
        # Send a write to a bank if the read to that bank isn't happening (unless you can do both)
        # further, we make sure there is a queued write, or current buffer is
        # full and there is an incoming push
        # and there is stuff in memory to be read
        self._wen_to_strg[idx] = ((~self._ren_to_strg[idx] | kts.const(int(self.rw_same_cycle), 1)) &
                                  (self._queued_write[idx] |  # Already queued a write...
                                  (((self._front_occ == self.fw_int) & self._push &  # Grossness
                                    (~self._front_pop)) &
                                   (self._curr_bank_wr == idx))))

    # Technically the ren should be one hot TODO: make onehot assertion
    @always_comb
    def send_reads(self, idx):
        # Send a read to a bank if there is only read_delay items left and it is being read
        self._ren_to_strg[idx] = ((((self._back_occ == self.read_delay) | self._fw_is_1)) &
                                  (self._curr_bank_rd == idx) &  # Proper bank
                                  (self._pop | ((self._back_occ == 0) & (self._back_num_load == 0))) &
                                  ((self._num_words_mem > 1) | ((self._num_words_mem == 1) & ~self._back_pl)))

    # Num load to back is 0 unless we are doing a parallel load
    # in which case we can count it out
    @always_comb
    def set_back_num_load(self):
        if self._back_pl:
            self._back_num_load = kts.ternary(self._pop,
                                              kts.const(self.fw_int - 1, self._back_num_load.width),
                                              kts.const(self.fw_int, self._back_num_load.width))
        else:
            self._back_num_load = 0

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_back_occ(self):
        if ~self._rst_n:
            self._back_occ = 0
        elif self._back_pl:
            if self._back_num_load == 0:
                self._back_occ = self._back_push.extend(self._back_occ.width)
            else:
                self._back_occ = self._back_num_load
        elif self._back_push & ~self._back_pop:
            self._back_occ = self._back_occ + 1
        elif ~self._back_push & self._back_pop:
            self._back_occ = self._back_occ - 1

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_front_occ(self):
        if ~self._rst_n:
            self._front_occ = 0
        elif self._front_par_read:
            if self._front_push:
                self._front_occ = 1
            else:
                self._front_occ = 0
        elif self._front_push & ~self._front_pop:
            self._front_occ = self._front_occ + 1
        elif ~self._front_push & self._front_pop:
            self._front_occ = self._front_occ - 1

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_parallel_ld_delay_1(self):
        if ~self._rst_n:
            self._ren_delay = 0
        else:
            self._ren_delay = self._ren_to_strg.r_or()

    @always_comb
    def set_front_par_read(self):
        # If the front is full and not being drained, need to write out to write queue or memory
        self._front_par_read = (self._front_occ == self.fw_int) & self._push & ~self._front_pop

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_num_words_mem(self):
        if ~self._rst_n:
            self._num_words_mem = 0
        elif ~self._back_pl & self._front_par_read:
            self._num_words_mem = self._num_words_mem + 1
        elif self._back_pl & ~self._front_par_read:
            self._num_words_mem = self._num_words_mem - 1

    # This signal drives the automatic passthrough to the back end
    # from the front end - bypassing the storage altogether
    # ----------------------------------------------------------
    # Basically check the nothing is in memory and
    # the back fifo isn't actually using a recent read
    # Also make sure that either the back fifo isn't full or it's being read from
    @always_comb
    def set_front_pop(self):
        self._front_pop = (((self._num_words_mem == 0) | ((self._num_words_mem == 1) & self._back_pl)) &
                           # You can pop the front if the memory read is being entirely bypassed (fw == 1)
                           (~self._back_pl | (self._back_pl & (self._back_num_load == 0))) &
                           (~(self._back_occ == self.fw_int) | self._pop) &
                           (~(self._front_occ == 0) | self._push))

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_write_queue(self, idx):
        if ~self._rst_n:
            self._write_queue[idx] = 0
            self._queued_write[idx] = 0
        # If the parallel read is happening, but the write is not
        # going through, then you must have a conflict and be queueing
        # Only write the queue for the current bank
        elif self._front_par_read & ~self._wen_to_strg[idx] & (self._curr_bank_wr == idx):
            self._write_queue[idx] = self._front_combined
            self._queued_write[idx] = 1
        elif self._wen_to_strg[idx]:
            self._queued_write[idx] = 0

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_curr_bank_wr(self):
        if ~self._rst_n:
            self._curr_bank_wr = 0
        elif self._front_par_read:
            if self._curr_bank_wr == self.banks - 1:
                self._curr_bank_wr = 0
            else:
                self._curr_bank_wr = self._curr_bank_wr + 1

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_curr_bank_rd(self):
        if ~self._rst_n:
            self._curr_bank_rd = 0
        elif self._ren_to_strg.r_or():
            if self._curr_bank_rd == self.banks - 1:
                self._curr_bank_rd = 0
            else:
                self._curr_bank_rd = self._curr_bank_rd + 1

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_prev_bank_rd(self):
        if ~self._rst_n:
            self._prev_bank_rd = 0
        # elif self._front_par_read:
        else:
            self._prev_bank_rd = self._curr_bank_rd

    @always_comb
    def set_num_items(self):
        self._num_items = ((self._num_words_mem * self.fw_int).extend(self.data_width) +
                           self._front_occ.extend(self.data_width) +
                           self._back_occ.extend(self.data_width))

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_wen_addr(self, idx):
        if ~self._rst_n:
            self._wen_addr[idx] = 0
        elif self._wen_to_strg[idx]:
            self._wen_addr[idx] = self._wen_addr[idx] + 1

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_ren_addr(self, idx):
        if ~self._rst_n:
            self._ren_addr[idx] = 0
        elif self._ren_to_strg[idx]:
            self._ren_addr[idx] = self._ren_addr[idx] + 1


if __name__ == "__main__":
    stg_dut = StrgFIFO()
    verilog(stg_dut, filename="strg_fifo.sv",
            additional_passes={"lift config regs": lift_config_reg})



from kratos import *
from lake.modules.aggregator import Aggregator
from lake.modules.addr_gen import AddrGen
from lake.attributes.config_reg_attr import ConfigRegAttr
from lake.passes.passes import lift_config_reg
import kratos as kts


class InputAddrCtrl(Generator):
    '''
    Input addressing control from the aggregation buffers to the SRAM
    '''
    def __init__(self,
                 interconnect_input_ports=2,
                 mem_depth=32,
                 num_tiles=1,
                 banks=1,
                 iterator_support=6,
                 address_width=5,
                 data_width=16,
                 fetch_width=16,
                 multiwrite=1,
                 strg_wr_ports=2,
                 config_width=16):
        super().__init__("input_addr_ctrl", debug=True)

        assert multiwrite >= 1, "Multiwrite must be at least 1..."

        self.interconnect_input_ports = interconnect_input_ports
        self.mem_depth = mem_depth
        self.num_tiles = num_tiles
        self.banks = banks
        self.iterator_support = iterator_support
        self.address_width = address_width
        self.port_sched_width = max(1, clog2(self.interconnect_input_ports))
        self.data_width = data_width
        self.fetch_width = fetch_width
        self.fw_int = int(self.fetch_width / self.data_width)
        self.multiwrite = multiwrite
        self.strg_wr_ports = strg_wr_ports
        self.config_width = config_width

        self.mem_addr_width = clog2(self.num_tiles * self.mem_depth)
        if self.banks > 1:
            self.bank_addr_width = clog2(self.banks)
        else:
            self.bank_addr_width = 0
        self.address_width = self.mem_addr_width + self.bank_addr_width

        # Clock and Reset
        self._clk = self.clock("clk")
        self._rst_n = self.reset("rst_n")

        # Inputs
        # phases = [] TODO
        # Take in the valid and data and attach an address + direct to a port
        self._valid_in = self.input("valid_in", self.interconnect_input_ports)
        self._wen_en = self.input("wen_en", self.interconnect_input_ports)

        self._wen_en_saved = self.var("wen_en_saved", self.interconnect_input_ports)

        self._data_in = self.input("data_in",
                                   self.data_width,
                                   size=(self.interconnect_input_ports,
                                         self.fw_int),
                                   explicit_array=True,
                                   packed=True)

        self._data_in_saved = self.var("data_in_saved",
                                       self.data_width,
                                       size=(self.interconnect_input_ports,
                                             self.fw_int),
                                       explicit_array=True,
                                       packed=True)

        # Outputs
        self._wen = self.output("wen_to_sram", self.strg_wr_ports,
                                size=self.banks,
                                explicit_array=True,
                                packed=True)

        wen_full_size = (self.interconnect_input_ports,
                         self.multiwrite)
        self._wen_full = self.var("wen_full",
                                  self.banks,
                                  size=wen_full_size,
                                  explicit_array=True,
                                  packed=True)

        self._wen_reduced = self.var("wen_reduced",
                                     self.banks,
                                     size=self.interconnect_input_ports,
                                     explicit_array=True,
                                     packed=True)

        self._wen_reduced_saved = self.var("wen_reduced_saved",
                                           self.banks,
                                           size=self.interconnect_input_ports,
                                           explicit_array=True,
                                           packed=True)

        self._addresses = self.output("addr_out",
                                      self.mem_addr_width,
                                      size=(self.banks,
                                            self.strg_wr_ports),
                                      explicit_array=True,
                                      packed=True)

        self._data_out = self.output("data_out",
                                     self.data_width,
                                     size=(self.banks,
                                           self.strg_wr_ports,
                                           self.fw_int),
                                     explicit_array=True,
                                     packed=True)

        self._port_out_exp = self.var("port_out_exp", self.interconnect_input_ports,
                                      size=self.banks,
                                      explicit_array=True,
                                      packed=True)

        self._port_out = self.output("port_out", self.interconnect_input_ports)

        self._counter = self.var("counter", self.port_sched_width)

        # Wire to port out
        for i in range(self.interconnect_input_ports):
            new_tmp = []
            for j in range(self.banks):
                new_tmp.append(self._port_out_exp[j][i])
            self.wire(self._port_out[i], kts.concat(*new_tmp).r_or())

        self._done = self.var("done", self.strg_wr_ports,
                              size=self.banks,
                              explicit_array=True,
                              packed=True)

        # LOCAL VARS
        self._local_addrs = self.var("local_addrs",
                                     self.address_width,
                                     size=(self.interconnect_input_ports,
                                           self.multiwrite),
                                     packed=True,
                                     explicit_array=True)

        self._local_addrs_saved = self.var("local_addrs_saved",
                                           self.address_width,
                                           size=(self.interconnect_input_ports,
                                                 self.multiwrite),
                                           packed=True,
                                           explicit_array=True)

        for i in range(self.interconnect_input_ports):
            for j in range(self.banks):
                concat_ports = []
                for k in range(self.multiwrite):
                    concat_ports.append(self._wen_full[i][k][j])
                self.wire(self._wen_reduced[i][j], kts.concat(*concat_ports).r_or())

        if self.banks == 1 and self.interconnect_input_ports == 1:
            self.wire(self._wen_full[0][0][0], self._valid_in)
        elif self.banks == 1 and self.interconnect_input_ports > 1:
            self.add_code(self.set_wen_single)
        else:
            self.add_code(self.set_wen_mult)

        # MAIN
        # Iterate through all banks to priority decode the wen
        self.add_code(self.decode_out_lowest)
        # Also set the write ports on the storage
        if self.strg_wr_ports > 1:
            self._idx_cnt = self.var("idx_cnt", 8,
                                     size=(self.banks,
                                           self.strg_wr_ports - 1),
                                     explicit_array=True,
                                     packed=True)
            for i in range(self.strg_wr_ports - 1):
                self.add_code(self.decode_out_alt, idx=i + 1)

        # Now we should instantiate the child address generators
        # (1 per input port) to send to the sram banks
        for i in range(self.interconnect_input_ports):
            self.add_child(f"address_gen_{i}", AddrGen(iterator_support=self.iterator_support,
                                                       config_width=self.config_width),
                           clk=self._clk,
                           rst_n=self._rst_n,
                           clk_en=const(1, 1),
                           flush=const(0, 1),
                           step=self._valid_in[i])

        # Need to check that the address falls into the bank for implicit banking

        # Then, obey the input schedule to send the proper Aggregator to the output
        # The wen to sram should be that the valid for the selected port is high
        # Do the same thing for the output address
        assert self.multiwrite <= self.banks and self.multiwrite > 0,\
            "Multiwrite should be between 1 and banks"
        if self.multiwrite > 1:
            size = (self.interconnect_input_ports, self.multiwrite - 1)
            self._offsets_cfg = self.input("offsets_cfg",
                                           self.address_width,
                                           size=size,
                                           packed=True,
                                           explicit_array=True)
            doc = "These offsets provide the ability to write to multiple banks explicitly"
            self._offsets_cfg.add_attribute(ConfigRegAttr(doc))
        self.add_code(self.set_multiwrite_addrs)

        # to handle multiple input ports going to fewer SRAM write ports
        self.add_code(self.set_int_ports_counter)
        self.add_code(self.save_mult_int_signals)

    # Update the pointer and mux for input and output schedule
    # Now, obey the input schedule to send to the proper SRAM bank

    @always_comb
    def set_wen_mult(self):
        for i in range(self.interconnect_input_ports):
            for j in range(self.multiwrite):
                for k in range(self.banks):
                    self._wen_full[i][j][k] = 0
                    if (self._valid_in[i]):
                        if(self._local_addrs[i][j][self.mem_addr_width + self.bank_addr_width - 1,
                                                   self.mem_addr_width] == k):
                            self._wen_full[i][j][k] = 1

    @always_comb
    def set_wen_single(self):
        for i in range(self.interconnect_input_ports):
            for j in range(self.multiwrite):
                for k in range(self.banks):
                    self._wen_full[i][j][k] = 0
                    if (self._valid_in[i]):
                        self._wen_full[i][j][k] = 1

    @always_comb
    def decode_out_lowest(self):
        for i in range(self.banks):
            self._wen[i][0] = 0
            self._done[i][0] = 0
            self._port_out_exp[i] = 0
            self._data_out[i][0] = 0
            self._addresses[i][0] = 0
            if ~self._done[i][0]:
                # In comb block, directly write data, addr, wen_en from first interconnect input port
                # If this input port is active on this bank...
                if self._wen_reduced[0][i]:
                    self._done[i][0] = 1
                    # This should only go through if the wen_en is on...
                    self._wen[i][0] = self._wen_en[0]
                    self._port_out_exp[i][0] = 1
                    self._data_out[i][0] = self._data_in[0]
                    self._addresses[i][0] = self._local_addrs[0][0][self.mem_addr_width - 1, 0]
                # Check to write data from other interconnect input ports (other than first)
                # If this input port is active on this bank...
                elif self.interconnect_input_ports > 1:
                    if self._wen_reduced_saved[self._counter + 1][i]:
                        self._done[i][0] = 1
                        # This should only go through if the wen_en is on...
                        self._wen[i][0] = self._wen_en_saved[self._counter + 1]
                        self._port_out_exp[i][self._counter + 1] = 1
                        self._data_out[i][0] = self._data_in_saved[self._counter + 1]
                        self._addresses[i][0] = \
                            self._local_addrs_saved[self._counter + 1][0][self.mem_addr_width - 1, 0]

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def save_mult_int_signals(self):
        if ~self._rst_n:
            self._wen_en_saved = 0
            self._wen_reduced_saved = 0
            self._data_in_saved = 0
            self._local_addrs_saved = 0
        # keep track of data from other interconnect input ports since we only are able
        # to write first interconnect input port at first
        elif self._wen_en.r_or():
            self._wen_en_saved = self._wen_en
            self._wen_reduced_saved = self._wen_reduced
            self._data_in_saved = self._data_in
            self._local_addrs_saved = self._local_addrs
        elif self.interconnect_input_ports == 1:
            self._wen_en_saved = 0
            self._wen_reduced_saved = 0
            self._data_in_saved = 0
            self._local_addrs_saved = 0
        # the data/wen from first interconnect input port is directly written, so
        # we do not have to resend that to SRAM
        elif self._counter == self.interconnect_input_ports - 2:
            self._wen_en_saved = 0
            self._wen_reduced_saved = 0
            self._data_in_saved = 0
            self._local_addrs_saved = 0

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_int_ports_counter(self):
        if ~self._rst_n:
            self._counter = 0
        elif self.interconnect_input_ports == 1:
            self._counter = 0
        # the data/wen from first interconnect input port is directly written, so
        # we do not have to resend that to SRAM
        elif self._counter == self.interconnect_input_ports - 2:
            self._counter = 0
        elif self._wen_en_saved.r_or():
            self._counter = self._counter + 1

    @always_comb
    def decode_out_alt(self, idx):
        for i in range(self.banks):
            self._wen[i][idx] = 0
            self._done[i][idx] = 0
            self._data_out[i][idx] = 0
            self._addresses[i][idx] = 0
            self._idx_cnt[i][idx - 1] = 0
            for j in range(self.interconnect_input_ports):
                if ~self._done[i][idx]:
                    if self._wen_reduced[j][i] & (self._idx_cnt[i][idx - 1] == idx):
                        # self._wen[i][idx] = 1
                        self._wen[i][idx] = self._wen_en[j]
                        self._done[i][idx] = 1
                        self._data_out[i][idx] = self._data_in[j]
                        self._addresses[i][idx] = self._local_addrs[j][0][self.mem_addr_width - 1, 0]
                    self._idx_cnt[i][idx - 1] = self._idx_cnt[i][idx - 1] + 1

    @always_comb
    def set_multiwrite_addrs(self):
        for i in range(self.interconnect_input_ports):
            self._local_addrs[i][0] = self[f"address_gen_{i}"].ports.addr_out[self.address_width - 1, 0]
        for i in range(self.multiwrite - 1):
            for j in range(self.interconnect_input_ports):
                self._local_addrs[j][i + 1] = self._local_addrs[j][0] + self._offsets_cfg[j][i]


if __name__ == "__main__":
    # iac_dut = InputAddrCtrl(interconnect_input_ports=2,
    #                        data_width=16,
    #                        mem_depth=512,
    #                        banks=4,
    #                        iterator_support=6,
    #                        max_port_schedule=64,
    #                        address_width=16,
    #                        multiwrite=2)
    iac_dut = InputAddrCtrl()
    verilog(iac_dut, filename="input_addr_ctrl.sv",
            additional_passes={"lift config regs": lift_config_reg})



import kratos
from kratos import Attribute, verilog
from kratos import *
from lake.attributes.config_reg_attr import ConfigRegAttr
from lake.attributes.sram_port_attr import SRAMPortAttr
from lake.attributes.formal_attr import FormalAttr, FormalSignalConstraint
import _kratos


# this is a pass
def lift_config_reg(generator):
    # if not hasattr(generator, "lifted"):
    #     setattr(generator, "lifted", False)
    # if generator.lifted:
    #     print("WARNING: LAKE PASSES: LIFT_CONFIG_REGS: Configuration registers have already been lifted...")
    #     print("WARNING: LAKE PASSES: LIFT_CONFIG_REGS: Exiting pass early...")
    #     return
    class ConfigRegLiftVisitor(IRVisitor):
        def __init__(self):
            IRVisitor.__init__(self)

        def visit(self, node):
            if isinstance(node, _kratos.Generator):
                ports_ = node.get_port_names()
                for port_name in ports_:
                    curr_port = node.get_port(port_name)
                    attrs = curr_port.find_attribute(lambda a: isinstance(a, ConfigRegAttr))
                    annotation_attr = curr_port.find_attribute(lambda a: isinstance(a, FormalAttr))
                    if port_name is "mode":
                        print("Found mode...")
                        print(attrs)
                    if len(attrs) != 1:
                        continue

                    cr_attr = attrs[0]
                    doc = cr_attr.get_documentation()
                    # need to wire it to the instances parent and up
                    gen = node
                    parent_gen = gen.parent_generator()
                    child_port = curr_port
                    child_gen = gen
                    top_lvl_cfg = parent_gen is None
                    while parent_gen is not None:
                        # create a port based on the target's definition
                        new_name = child_gen.instance_name + "_" + child_port.name
                        p = parent_gen.port(child_port, new_name, False)
                        parent_gen.wire(child_port, p)
                        # move up the hierarchy
                        child_port = p
                        child_gen = parent_gen
                        parent_gen = parent_gen.parent_generator()
                    # Only add the attribute if this is a newly created port, not a top-level cfg reg
                    if top_lvl_cfg is False:
                        child_port_cra = ConfigRegAttr()
                        child_port_cra.set_documentation(doc)
                        child_port.add_attribute(child_port_cra)
                        if annotation_attr:
                            ann_att = annotation_attr[0]
                            annot_type = ann_att.get_formal_ann()
                            child_port.add_attribute(FormalAttr(f"{child_port}", annot_type))

    v = ConfigRegLiftVisitor()
    v.visit_root(generator)
    # Make sure we don't lift twice
    # generator.lifted = True


# Inputs:
#   - use_sram_stub: if stub is not being used, we are using external provided
#       sram macro and should replace port names accordingly
#   - sram_macro_info: information about sram macro, including port names to
#       be replaced
def change_sram_port_names(use_sram_stub, sram_macro_info):

    def change_sram_port_names_wrapper(generator):

        class SRAMPortNames(IRVisitor):
            def __init__(self, use_sram_stub, ports):
                IRVisitor.__init__(self)
                self.use_sram_stub = use_sram_stub
                self.sram_macro_info = sram_macro_info

            def visit(self, node):
                if self.use_sram_stub:
                    return
                if isinstance(node, Port):
                    attrs = node.find_attribute(lambda a: isinstance(a, SRAMPortAttr))
                    if len(attrs) != 1:
                        return
                    if node.name == "sram_addr":
                        node.name = self.sram_macro_info.addr_port_name
                    elif node.name == "sram_cen":
                        node.name = self.sram_macro_info.ce_port_name
                    elif node.name == "sram_clk":
                        node.name = self.sram_macro_info.clk_port_name
                    elif node.name == "sram_data_in":
                        node.name = self.sram_macro_info.data_in_port_name
                    elif node.name == "sram_data_out":
                        node.name = self.sram_macro_info.data_out_port_name
                    elif node.name == "sram_wen":
                        node.name = self.sram_macro_info.wen_port_name
                    elif node.name == "sram_wtsel":
                        node.name = self.sram_macro_info.wtsel_port_name
                    elif node.name == "sram_rtsel":
                        node.name = self.sram_macro_info.rtsel_port_name

        v = SRAMPortNames(use_sram_stub, sram_macro_info)
        v.visit_root(generator)

    return change_sram_port_names_wrapper



from kratos import *


class SRAMControl(Generator):
    def __init__(self,
                 data_width,
                 banks,
                 mem_depth):
        super().__init__("sram_control")

        # PORT DEFS: begin
        self.data_width = data_width
        self.banks = banks
        self.mem_depth = mem_depth
        self.mem_addr_width = clog2(self.mem_depth)

        self._clk = self.clock("clk")
        self._clk_en = self.input("clk_en", 1)
        self._reset = self.reset("reset")
        self._flush = self.input("flush", 1)

        self._data_in = self.input("data_in", data_width)
        self._wen = self.input("wen", 1)
        self._data_out = self.output("data_out", data_width)
        self._ren = self.input("ren", 1)
        self._addr_in = self.input("addr_in", 16)

        self._sram_to_mem_data = self.output("sram_to_mem_data",
                                             self.data_width,
                                             size=self.banks,
                                             explicit_array=True,
                                             packed=True)
        self._sram_to_mem_cen = self.output("sram_to_mem_cen",
                                            self.banks)
        self._sram_to_mem_wen = self.output("sram_to_mem_wen",
                                            self.banks)
        self._sram_to_mem_addr = self.output("sram_to_mem_addr",
                                             self.mem_addr_width,
                                             size=self.banks,
                                             explicit_array=True,
                                             packed=True)
        self._mem_to_sram_data = self.input("mem_to_sram_data",
                                            self.data_width,
                                            size=self.banks,
                                            explicit_array=True,
                                            packed=True)
        # PORT DEFS: end

        self.bank_addr_width = clog2(self.banks)
        self.full_addr = self.mem_addr_width + self.bank_addr_width

        # LOCAL SIGNALS: begin
        self._addr = self.var("addr", self.full_addr)
        self._bank_seld = self.var("bank_seld", self.banks)
        self._sram_to_mem_ren_reg = self.var("sram_to_mem_ren_reg", banks)
        self._data_out_reg = self.var("data_out_reg", self.data_width)
        # LOCAL SIGNALS: end

        # GENERATION LOGIC: begin
        self.wire(self._addr, self._addr_in[self.full_addr - 1, 0])
        for i in range(self.banks):
            self.wire(self._bank_seld[i],
                      self._addr[self.full_addr - 1,
                                 self.full_addr - self.bank_addr_width] == i)
            self.wire(self._sram_to_mem_data[i], self._data_in)
            self.wire(self._sram_to_mem_cen[i], self._bank_seld[i] & (self._wen | self._ren))
            self.wire(self._sram_to_mem_wen[i], self._bank_seld[i] & self._wen)
            self.wire(self._sram_to_mem_addr[i], self._addr[self.mem_addr_width - 1, 0])

        self.bank_sel_comb = CombinationalCodeBlock(self)
        out_switch = SwitchStmt(self._sram_to_mem_ren_reg)
        for i in range(self.banks):
            out_switch.case_(2 ** i, self._data_out.assign(self._mem_to_sram_data[i]))
        out_switch.case_(None, self._data_out.assign(self._data_out_reg))
        self.bank_sel_comb.add_stmt(out_switch)
        self.add_code(self.data_out_reg_update)
        self.add_code(self.ren_reg_update)
        # GENERATION LOGIC: end

    @always((posedge, "clk"), (posedge, "reset"))
    def data_out_reg_update(self):
        if self._reset:
            self._data_out_reg = 0
        elif self._clk_en:
            if self._flush:
                self._data_out_reg = 0
            else:
                self._data_out_reg = self._data_out

    @always((posedge, "clk"), (posedge, "reset"))
    def ren_reg_update(self):
        if self._reset:
            self._sram_to_mem_ren_reg = 0
        elif self._clk_en:
            if self._flush:
                self._sram_to_mem_ren_reg = 0
            else:
                self._sram_to_mem_ren_reg = self._bank_seld & concat(*[self._ren
                                                                       for i in range(self.banks)])


# Python main guard
if __name__ == "__main__":
    dut = SRAMControl(16, 2, 512)
    verilog(dut, filename="sram_control.sv", check_active_high=False)



import kratos
from kratos import *
from math import log
from lake.attributes.config_reg_attr import ConfigRegAttr


class TransposeBuffer(Generator):
    def __init__(self,
                 # number of bits in a word
                 word_width,
                 # number of words that can be sotred at an address in SRAM
                 # note fetch_width must be powers of 2
                 fetch_width,
                 # total number of transpose buffers
                 num_tb,
                 # height of this particular transpose buffer
                 max_tb_height,
                 # maximum value for range parameters in nested for loop
                 # (and as a result, maximum length of indices input vector)
                 # specifying inner for loop values for output column
                 # addressing
                 max_range,
                 max_range_inner,
                 max_stride,
                 tb_iterator_support):
        super().__init__("transpose_buffer", debug=True)

        #########################
        # GENERATION PARAMETERS #
        #########################

        self.word_width = word_width
        self.fetch_width = fetch_width
        self.num_tb = num_tb
        self.max_tb_height = max_tb_height
        self.max_range = max_range
        self.max_range_inner = max_range_inner
        self.max_stride = max_stride
        self.tb_iterator_support = tb_iterator_support

        ##################################
        # BITS FOR GENERATION PARAMETERS #
        ##################################

        self.fetch_width_bits = max(1, clog2(self.fetch_width))
        self.num_tb_bits = max(1, clog2(self.num_tb))
        self.max_range_bits = max(1, clog2(self.max_range))
        self.max_range_inner_bits = max(1, clog2(self.max_range_inner))
        self.max_stride_bits = max(1, clog2(self.max_stride))
        self.tb_col_index_bits = 2 * max(self.fetch_width_bits, self.num_tb_bits) + 1
        self.max_tb_height_bits2 = max(1, clog2(2 * self.max_tb_height))
        self.max_tb_height_bits = max(1, clog2(self.max_tb_height))
        self.tb_iterator_support_bits = max(1, clog2(self.tb_iterator_support) + 1)
        self.max_range_stride_bits2 = max(2 * self.max_range_bits, 2 * self.max_stride_bits)

        ##########
        # INPUTS #
        ##########

        self.clk = self.clock("clk")
        # active low asynchronous reset
        self.rst_n = self.reset("rst_n", 1)

        # data input from SRAM
        if self.fetch_width == 1:
            self.input_data = self.input("input_data", self.word_width)
        else:
            self.input_data = self.input("input_data",
                                         width=self.word_width,
                                         size=self.fetch_width,
                                         packed=True)
        # valid indicating whether data input from SRAM is valid and
        # should be stored in transpose buffer
        self.valid_data = self.input("valid_data", 1)
        self.ack_in = self.input("ack_in", 1)
        self.ren = self.input("ren", 1)

        self.mem_valid_data = self.input("mem_valid_data", 1)

        ###########################
        # CONFIGURATION REGISTERS #
        ###########################

        # the range of the outer for loop in nested for loop for output
        # column address generation
        self.range_outer = self.input("range_outer", self.max_range_bits)
        self.range_outer.add_attribute(ConfigRegAttr("Outer range for output for loop pattern"))

        # the range of the inner for loop in nested for loop for output
        # column address generation
        self.range_inner = self.input("range_inner", self.max_range_inner_bits)
        self.range_inner.add_attribute(ConfigRegAttr("Inner range for output for for loop pattern"))

        # stride for the given application
        self.stride = self.input("stride", self.max_stride_bits)
        self.stride.add_attribute(ConfigRegAttr("Application stride"))

        self.tb_height = self.input("tb_height", self.max_tb_height_bits)
        self.tb_height.add_attribute(ConfigRegAttr("Transpose Buffer height"))

        self.dimensionality = self.input("dimensionality", self.tb_iterator_support_bits)
        self.dimensionality.add_attribute(ConfigRegAttr("Transpose Buffer dimensionality"))

        # specifies inner for loop values for output column
        # addressing
        self.indices = self.input("indices",
                                  width=clog2(2 * self.num_tb * self.fetch_width),
                                  # the length of indices is equal to range_inner,
                                  # so the maximum possible size for self.indices
                                  # is the maximum value of range_inner, which is
                                  # self.max_range_inner
                                  size=self.max_range_inner,
                                  explicit_array=True,
                                  packed=True)
        self.indices.add_attribute(ConfigRegAttr("Output indices for for loop pattern"))

        # offset to start output address if we're starting in the middle of a wider
        # fetch width word for example
        self.starting_addr = self.input("starting_addr", self.fetch_width_bits)
        self.starting_addr.add_attribute(ConfigRegAttr("TB starting address"))

        ###########
        # OUTPUTS #
        ###########

        self.col_pixels = self.output("col_pixels",
                                      width=self.word_width,
                                      size=self.max_tb_height,
                                      packed=True,
                                      explicit_array=True)
        self.output_valid = self.output("output_valid", 1)
        self.rdy_to_arbiter = self.output("rdy_to_arbiter", 1)

        ###################
        # LOCAL VARIABLES #
        ###################

        # transpose buffer
        if self.fetch_width == 1:
            self.tb = self.var("tb",
                               width=self.word_width,
                               size=2 * self.max_tb_height,
                               packed=True)

        else:
            self.tb = self.var("tb",
                               width=self.word_width,
                               size=[2 * self.max_tb_height, self.fetch_width],
                               packed=True)

        self.tb_valid = self.var("tb_valid", 2 * self.max_tb_height)

        self.index_outer = self.var("index_outer", self.max_range_bits)
        self.index_inner = self.var("index_inner", self.max_range_inner_bits)

        self.input_buf_index = self.var("input_buf_index", 1)
        self.out_buf_index = self.var("out_buf_index", 1)
        self.switch_out_buf = self.var("switch_out_buf", 1)
        self.switch_next_line = self.var("switch_next_line", 1)
        self.row_index = self.var("row_index", self.max_tb_height_bits)
        self.input_index = self.var("input_index", self.max_tb_height_bits2)

        self.output_index_abs = self.var("output_index_abs", self.max_range_stride_bits2)

        if self.fetch_width != 1:
            self.output_index_long = self.var("output_index_long", self.max_range_stride_bits2)
            self.output_index = self.var("output_index", self.fetch_width_bits)

        self.indices_index_inner = self.var("indices_index_inner",
                                            clog2(2 * self.num_tb * self.fetch_width))
        self.curr_out_start = self.var("curr_out_start", self.max_range_stride_bits2)

        self.start_data = self.var("start_data", 1)
        self.old_start_data = self.var("old_start_data", 1)

        self.pause_tb = self.var("pause_tb", 1)
        self.pause_output = self.var("pause_output", 1)

        self.on_next_line = self.var("on_next_line", 1)

        self.mask_valid = self.var("mask_valid", 1)

        self.pause_tbinv = self.var("pause_tbinv", 1)
        self.rdy_to_arbiterinv = self.var("rdy_to_arbiterinv", 1)
        self.out_buf_indexinv = self.var("out_buf_indexinv", 1)

        ##########################
        # SEQUENTIAL CODE BLOCKS #
        ##########################

        self.add_code(self.set_index_outer)
        self.add_code(self.set_index_inner)
        self.add_code(self.set_pause_tb)
        self.add_code(self.set_row_index)
        self.add_code(self.set_input_buf_index)
        self.add_code(self.input_to_tb)
        self.add_code(self.output_from_tb)
        self.add_code(self.set_output_valid)
        self.add_code(self.set_out_buf_index)
        self.add_code(self.set_rdy_to_arbiter)
        self.add_code(self.set_start_data)
        self.add_code(self.set_curr_out_start)
        if self.fetch_width != 1:
            self.add_code(self.set_output_index)
        self.add_code(self.set_old_start_data)
        self.add_code(self.set_on_next_line)

        #############################
        # COMBINATIONAL CODE BLOCKS #
        #############################

        self.add_code(self.set_pause_output)
        self.add_code(self.set_input_index)
        self.add_code(self.set_tb_out_indices)
        self.add_code(self.set_switch_out_buf)
        self.add_code(self.set_switch_next_line)
        if self.fetch_width != 1:
            self.add_code(self.set_output_index_long)
        self.add_code(self.set_mask_valid)
        self.add_code(self.set_invs)

    # get output loop iterators
    # set pause_tb signal to pause input/output depending on
    # output loop iterator values and valid_data
    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_index_outer(self):
        if ~self.rst_n:
            self.index_outer = 0
        elif self.dimensionality == 0:
            self.index_outer = 0
        elif (self.dimensionality == 1) | \
                ((self.dimensionality == 2) & (self.index_inner == self.range_inner - 1)):
            if ~self.pause_output:
                if (self.index_outer == self.range_outer - 1):
                    self.index_outer = 0
                else:
                    self.index_outer = self.index_outer + 1

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_index_inner(self):
        if ~self.rst_n:
            self.index_inner = 0
        elif self.dimensionality <= 1:
            self.index_inner = 0
        elif ~self.pause_output:
            if self.index_inner == self.range_inner - 1:
                self.index_inner = 0
            else:
                self.index_inner = self.index_inner + 1

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_pause_tb(self):
        if ~self.rst_n:
            self.pause_tbinv = 0
        elif self.dimensionality == 0:
            self.pause_tbinv = 0
        elif (self.index_outer == self.range_outer - 1) & \
                ((self.dimensionality == 1) |
                    ((self.dimensionality == 2) &
                        (self.index_inner == self.range_inner - 1))):
            if ~self.pause_output:
                # do not need to pause tb if
                # - rdy is not high as we currently have valid data in tb that we can
                # keep outputting
                # - we get valid data as we can output that
                if ~self.rdy_to_arbiter | self.valid_data:
                    self.pause_tbinv = 1
                else:
                    self.pause_tbinv = 0
        elif self.pause_tb:
            self.pause_tbinv = self.valid_data

    @always_comb
    def set_invs(self):
        self.pause_tb = ~self.pause_tbinv
        self.out_buf_index = ~self.out_buf_indexinv
        self.rdy_to_arbiter = ~self.rdy_to_arbiterinv

    @always_comb
    def set_pause_output(self):
        if self.pause_tb:
            self.pause_output = 1
        else:
            self.pause_output = ~self.ren

    # get index of row to fill in transpose buffer with input data
    # for one of the two buffers in double buffer
    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_row_index(self):
        if ~self.rst_n:
            self.row_index = 0
        elif self.dimensionality == 0:
            self.row_index = 0
        elif self.valid_data & self.row_index == self.tb_height - 1:
            self.row_index = 0
        elif self.valid_data:
            self.row_index = self.row_index + 1

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_input_buf_index(self):
        if ~self.rst_n:
            self.input_buf_index = 0
        elif self.valid_data & (self.row_index == self.tb_height - 1):
            self.input_buf_index = ~self.input_buf_index

    # for double buffer, get index of row to fill in transpose buffer
    # with input data
    @always_comb
    def set_input_index(self):
        if self.dimensionality == 0:
            self.input_index = 0
        elif self.input_buf_index:
            self.input_index = const(self.max_tb_height, self.max_tb_height_bits2) + \
                self.row_index.extend(self.max_tb_height_bits2)
        else:
            self.input_index = self.row_index.extend(self.max_tb_height_bits2)

    # input to transpose buffer
    @always_ff((posedge, "clk"))
    def input_to_tb(self):
        if self.valid_data:
            if self.dimensionality == 0:
                self.tb[self.input_index] = 0
                self.tb_valid[self.input_index] = 0
            else:
                self.tb[self.input_index] = self.input_data
                self.tb_valid[self.input_index] = self.mem_valid_data

    # get relative output column index from absolute output column index
    @always_comb
    def set_tb_out_indices(self):
        if self.dimensionality == 0:
            self.indices_index_inner = 0
            self.output_index_abs = 0
        elif self.dimensionality == 1:
            self.indices_index_inner = 0
            self.output_index_abs = self.index_outer.extend(self.max_range_stride_bits2) * \
                self.stride.extend(self.max_range_stride_bits2) \
                + self.starting_addr.extend(self.max_range_stride_bits2)
        else:
            self.indices_index_inner = self.indices[self.index_inner]
            self.output_index_abs = self.index_outer.extend(self.max_range_stride_bits2) * \
                self.stride.extend(self.max_range_stride_bits2) \
                + self.indices_index_inner.extend(self.max_range_stride_bits2) \
                + self.starting_addr.extend(self.max_range_stride_bits2)

    @always_comb
    def set_output_index_long(self):
        self.output_index_long = self.output_index_abs % fetch_width

    # output column from transpose buffer
    @always_comb
    def output_from_tb(self):
        for i in range(max_tb_height):
            self.col_pixels[i] = 0
            if i < self.tb_height:
                if self.dimensionality == 0:
                    self.col_pixels[i] = 0
                elif (self.out_buf_index ^ self.switch_out_buf):
                    if self.fetch_width == 1:
                        self.col_pixels[i] = self.tb[i]
                    else:
                        self.col_pixels[i] = self.tb[i][self.output_index]
                else:
                    if self.fetch_width == 1:
                        self.col_pixels[i] = self.tb[i + self.max_tb_height]
                    else:
                        self.col_pixels[i] = self.tb[i + self.max_tb_height][self.output_index]

    @always_comb
    def set_output_index(self):
        self.output_index = self.output_index_long[self.fetch_width_bits - 1, 0]

    @always_comb
    def set_switch_next_line(self):
        if ((self.index_outer == 0) & ~self.on_next_line &
                ((self.dimensionality == 1) |
                    ((self.dimensionality == 2) & (self.index_inner == 0)))):
            self.switch_next_line = 1
        else:
            self.switch_next_line = 0

    @always_comb
    def set_switch_out_buf(self):
        if self.switch_next_line | \
                (self.output_index_abs >= self.curr_out_start + self.fetch_width):
            self.switch_out_buf = 1
        else:
            self.switch_out_buf = 0

    # generates output valid and updates which buffer in double buffer to output from
    # appropriately
    @always_comb
    def set_output_valid(self):
        if self.dimensionality == 0:
            self.output_valid = 0
        elif self.pause_output:
            self.output_valid = 0
        else:
            self.output_valid = self.mask_valid

    @always_comb
    def set_mask_valid(self):
        if (self.out_buf_index ^ self.switch_out_buf):
            self.mask_valid = self.tb_valid[0]
        else:
            self.mask_valid = self.tb_valid[1]

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_out_buf_index(self):
        if ~self.rst_n:
            self.out_buf_indexinv = 0
        elif ~self.start_data:
            self.out_buf_indexinv = 0
        elif self.switch_out_buf:
            self.out_buf_indexinv = ~self.out_buf_indexinv

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_on_next_line(self):
        if ~self.rst_n:
            self.on_next_line = 0
        elif self.switch_next_line:
            self.on_next_line = 1
        elif (self.index_outer == self.range_outer - 1):
            if (self.dimensionality == 1) | \
                    ((self.dimensionality == 2) & (self.index_inner == self.range_inner - 1)):
                self.on_next_line = 0

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_curr_out_start(self):
        if ~self.rst_n:
            self.curr_out_start = 0
        elif self.dimensionality == 0:
            self.curr_out_start = 0
        elif self.switch_next_line:
            self.curr_out_start = 0
        elif (self.output_index_abs >= self.curr_out_start + self.fetch_width):
            self.curr_out_start = self.curr_out_start + self.fetch_width

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_start_data(self):
        if ~self.rst_n:
            self.start_data = 0
        elif self.valid_data & ~self.start_data:
            self.start_data = 1

    @always_ff((posedge, "clk"))
    def set_old_start_data(self):
        self.old_start_data = self.start_data

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_rdy_to_arbiter(self):
        if ~self.rst_n:
            self.rdy_to_arbiterinv = 0
        elif self.dimensionality == 0:
            self.rdy_to_arbiterinv = 1
        elif self.start_data & ~self.old_start_data:
            self.rdy_to_arbiterinv = 0
        elif self.switch_out_buf:
            self.rdy_to_arbiterinv = 0
        elif self.tb_height != 1:
            if self.row_index != self.tb_height - 1:
                self.rdy_to_arbiterinv = 0
        elif self.ack_in:
            self.rdy_to_arbiterinv = 1


if __name__ == "__main__":
    dut = TransposeBuffer(word_width=16,
                          fetch_width=4,
                          num_tb=1,
                          max_tb_height=1,
                          max_range=9,
                          max_range_inner=5,
                          max_stride=3,
                          tb_iterator_support=2)
    verilog(dut, filename="transpose_buffer.sv")



from kratos import *
from lake.attributes.config_reg_attr import ConfigRegAttr
from lake.passes.passes import lift_config_reg
import kratos as kts


class AppCtrl(Generator):
    '''
    Application Controller.
        This module holds the logic/expression of read/write sequencing at the application level.
        For example, a pure double buffering application needs a deterministic lockstep between sets
        of reads and writes. The constraint imposed on this application prevents the divided memory space
        from being self-conflicting.
        For a line buffer, we only need
    '''
    def __init__(self,
                 interconnect_input_ports,
                 interconnect_output_ports,
                 depth_width=16,
                 sprt_stcl_valid=False,
                 stcl_cnt_width=16,
                 stcl_iter_support=4):
        super().__init__("app_ctrl", debug=True)

        self.int_in_ports = interconnect_input_ports
        self.int_out_ports = interconnect_output_ports
        self.depth_width = depth_width
        self.sprt_stcl_valid = sprt_stcl_valid
        self.stcl_cnt_width = stcl_cnt_width
        self.stcl_iter_support = stcl_iter_support

        # Clock and Reset
        self._clk = self.clock("clk")
        self._rst_n = self.reset("rst_n")

        # IO
        self._wen_in = self.input("wen_in", self.int_in_ports)
        self._ren_in = self.input("ren_in", self.int_out_ports)

        self._ren_update = self.input("ren_update", self.int_out_ports)

        self._tb_valid = self.input("tb_valid", self.int_out_ports)

        self._valid_out_data = self.output("valid_out_data", self.int_out_ports)

        self._valid_out_stencil = self.output("valid_out_stencil", self.int_out_ports)

        # Send tb valid to valid out for now...

        if self.sprt_stcl_valid:
            # Add the config registers to watch
            self._ranges = self.input("ranges", self.stcl_cnt_width,
                                      size=self.stcl_iter_support,
                                      packed=True,
                                      explicit_array=True)
            self._ranges.add_attribute(ConfigRegAttr("Ranges of stencil valid generator"))

            self._threshold = self.input("threshold", self.stcl_cnt_width,
                                         size=self.stcl_iter_support,
                                         packed=True,
                                         explicit_array=True)
            self._threshold.add_attribute(ConfigRegAttr("Threshold of stencil valid generator"))

            self._dim_counter = self.var("dim_counter", self.stcl_cnt_width,
                                         size=self.stcl_iter_support,
                                         packed=True,
                                         explicit_array=True)

            self._update = self.var("update", self.stcl_iter_support)

            self.wire(self._update[0], const(1, 1))
            for i in range(self.stcl_iter_support - 1):
                self.wire(self._update[i + 1],
                          (self._dim_counter[i] == (self._ranges[i] - 1)) & self._update[i])

            for i in range(self.stcl_iter_support):
                self.add_code(self.dim_counter_update, idx=i)

            # Now we need to just compute stencil valid
            threshold_comps = [self._dim_counter[_i] >= self._threshold[_i] for _i in range(self.stcl_iter_support)]
            self.wire(self._valid_out_stencil[0], kts.concat(*threshold_comps).r_and())
            for i in range(self.int_out_ports - 1):
                # self.wire(self._valid_out_stencil[i + 1], 0)
                # for multiple ports
                self.wire(self._valid_out_stencil[i + 1], kts.concat(*threshold_comps).r_and())

        else:
            self.wire(self._valid_out_stencil, self._tb_valid)

        # Now gate the valid with stencil valid
        self.wire(self._valid_out_data, self._tb_valid & self._valid_out_stencil)
        self._wr_delay_state_n = self.var("wr_delay_state_n", self.int_out_ports)
        self._wen_out = self.output("wen_out", self.int_in_ports)
        self._ren_out = self.output("ren_out", self.int_out_ports)

        self._write_depth_wo = self.input("write_depth_wo", self.depth_width,
                                          size=self.int_in_ports,
                                          explicit_array=True,
                                          packed=True)
        self._write_depth_wo.add_attribute(ConfigRegAttr("Depth of writes"))

        self._write_depth_ss = self.input("write_depth_ss", self.depth_width,
                                          size=self.int_in_ports,
                                          explicit_array=True,
                                          packed=True)
        self._write_depth_ss.add_attribute(ConfigRegAttr("Depth of writes"))

        self._write_depth = self.var("write_depth", self.depth_width,
                                     size=self.int_in_ports,
                                     explicit_array=True,
                                     packed=True)

        for i in range(self.int_in_ports):
            self.wire(self._write_depth[i],
                      kts.ternary(self._wr_delay_state_n[i],
                                  self._write_depth_ss[i],
                                  self._write_depth_wo[i]))

        self._read_depth = self.input("read_depth", self.depth_width,
                                      size=self.int_out_ports,
                                      explicit_array=True,
                                      packed=True)
        self._read_depth.add_attribute(ConfigRegAttr("Depth of reads"))

        self._write_count = self.var("write_count", self.depth_width,
                                     size=self.int_in_ports,
                                     explicit_array=True,
                                     packed=True)

        self._read_count = self.var("read_count", self.depth_width,
                                    size=self.int_out_ports,
                                    explicit_array=True,
                                    packed=True)
        self._write_done = self.var("write_done", self.int_in_ports)
        self._write_done_ff = self.var("write_done_ff", self.int_in_ports)
        self._read_done = self.var("read_done", self.int_out_ports)
        self._read_done_ff = self.var("read_done_ff", self.int_out_ports)

        self.in_port_bits = max(1, kts.clog2(self.int_in_ports))
        self._input_port = self.input("input_port", self.in_port_bits,
                                      size=self.int_out_ports,
                                      explicit_array=True,
                                      packed=True)
        self._input_port.add_attribute(ConfigRegAttr("Relative input port for an output port"))

        self.out_port_bits = max(1, kts.clog2(self.int_out_ports))
        self._output_port = self.input("output_port", self.out_port_bits,
                                       size=self.int_in_ports,
                                       explicit_array=True,
                                       packed=True)
        self._output_port.add_attribute(ConfigRegAttr("Relative output port for an input port"))

        self._prefill = self.input("prefill", self.int_out_ports)
        self._prefill.add_attribute(ConfigRegAttr("Is the input stream prewritten?"))

        for i in range(self.int_out_ports):
            self.add_code(self.set_read_done, idx=i)
            if self.int_in_ports == 1:
                self.add_code(self.set_read_done_ff_one_wr, idx=i)
            else:
                self.add_code(self.set_read_done_ff, idx=i)

        # self._write_done_comb = self.var("write_done_comb", self.int_in_ports)
        for i in range(self.int_in_ports):
            self.add_code(self.set_write_done, idx=i)
            self.add_code(self.set_write_done_ff, idx=i)

        for i in range(self.int_in_ports):
            self.add_code(self.set_write_cnt, idx=i)
        for i in range(self.int_out_ports):
            if self.int_in_ports == 1:
                self.add_code(self.set_read_cnt_one_wr, idx=i)
            else:
                self.add_code(self.set_read_cnt, idx=i)

        for i in range(self.int_out_ports):
            if self.int_in_ports == 1:
                self.add_code(self.set_wr_delay_state_one_wr, idx=i)
            else:
                self.add_code(self.set_wr_delay_state, idx=i)

        self._read_on = self.var("read_on", self.int_out_ports)
        for i in range(self.int_out_ports):
            self.wire(self._read_on[i], self._read_depth[i].r_or())

        # If we have prefill enabled, we are skipping the initial delay step...
        self.wire(self._ren_out,
                  (self._wr_delay_state_n | self._prefill) & ~self._read_done_ff & self._ren_in &
                  self._read_on)
        self.wire(self._wen_out, ~self._write_done_ff & self._wen_in)

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_write_done_ff(self, idx):
        if ~self._rst_n:
            self._write_done_ff[idx] = 0
        elif self._write_done[idx] & self._read_done[self._output_port[idx]]:
            self._write_done_ff[idx] = 0
        elif self._write_done[idx]:
            self._write_done_ff[idx] = 1

    @always_comb
    def set_write_done(self, idx):
        self._write_done[idx] = (self._wen_in[idx] &
                                 (self._write_count[idx] == (self._write_depth[idx] - 1))) | \
            self._write_done_ff[idx]

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_write_cnt(self, idx):
        if ~self._rst_n:
            self._write_count[idx] = 0
        elif self._write_done[idx] & self._read_done[self._output_port[idx]]:
            self._write_count[idx] = 0
        elif self._wen_in[idx] & ~self._write_done_ff[idx]:
            self._write_count[idx] = self._write_count[idx] + 1

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_read_done_ff(self, idx):
        if ~self._rst_n:
            self._read_done_ff[idx] = 0
        elif self._write_done[self._input_port[idx]] & self._read_done[idx]:
            self._read_done_ff[idx] = 0
        elif self._read_done[idx]:
            self._read_done_ff[idx] = 1

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_read_done_ff_one_wr(self, idx):
        if ~self._rst_n:
            self._read_done_ff[idx] = 0
        elif self._write_done & self._read_done[idx]:
            self._read_done_ff[idx] = 0
        elif self._read_done[idx]:
            self._read_done_ff[idx] = 1

    @always_comb
    def set_read_done(self, idx):
        self._read_done[idx] = (self._ren_update[idx] & self._ren_in[idx] &
                                (self._read_count[idx] == (self._read_depth[idx] - 1))) | \
            self._read_done_ff[idx] | (~self._prefill[idx] & ~self._wr_delay_state_n[idx])

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_read_cnt(self, idx):
        if ~self._rst_n:
            self._read_count[idx] = 0
        elif self._write_done[self._input_port[idx]] & self._read_done[idx]:
            self._read_count[idx] = 0
        elif self._ren_in[idx] & self._ren_update[idx]:
            self._read_count[idx] = self._read_count[idx] + 1

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_read_cnt_one_wr(self, idx):
        if ~self._rst_n:
            self._read_count[idx] = 0
        elif self._write_done & self._read_done[idx]:
            self._read_count[idx] = 0
        elif self._ren_in[idx] & self._ren_update[idx]:
            self._read_count[idx] = self._read_count[idx] + 1

    # When we start up, there is no way to read data from storage.
    # We use this flag to gate read logic TODO : change to a separate counter
    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_wr_delay_state(self, idx):
        if ~self._rst_n:
            self._wr_delay_state_n[idx] = 0
        elif self._write_done[self._input_port[idx]]:
            self._wr_delay_state_n[idx] = 1

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_wr_delay_state_one_wr(self, idx):
        if ~self._rst_n:
            self._wr_delay_state_n[idx] = 0
        elif self._write_done:
            self._wr_delay_state_n[idx] = 1

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def dim_counter_update(self, idx):
        if ~self._rst_n:
            self._dim_counter[idx] = 0
        elif self._ren_in[0] & self._ren_update[0]:
            # if self._update[idx] & (idx < self._dimensionality):
            if self._update[idx]:
                if self._dim_counter[idx] == (self._ranges[idx] - 1):
                    self._dim_counter[idx] = 0
                else:
                    self._dim_counter[idx] = self._dim_counter[idx] + 1


if __name__ == "__main__":
    ac_dut = AppCtrl(interconnect_input_ports=1,
                     interconnect_output_ports=3)
    verilog(ac_dut, filename="app_ctrl.sv",
            additional_passes={"lift config regs": lift_config_reg})



from kratos import *


class PipeReg(Generator):
    def __init__(self,
                 data_width,
                 stages):
        super().__init__("pipe_reg")
        self._data_in = self.input("data_in", data_width)
        self._data_out = self.output("data_out", data_width)
        self._clk = self.clock("clk")
        self._rst_n = self.reset("rst_n")
        self._clk_en = self.input("clk_en", 1)

        self.stages = stages

        if stages > 1:
            self._reg_array = self.var("reg_array",
                                       data_width,
                                       size=stages,
                                       packed=True,
                                       explicit_array=True)
            self.add_code(self.stage_elab)
            self.add_code(self.set_out)
        elif stages == 1:
            self._reg_ = self.var("reg_", data_width)
            self.add_code(self.set_out_one_stage)
            self.wire(self._data_out, self._reg_)
        else:
            self.wire(self._data_out, self._data_in)

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def stage_elab(self):
        if ~self._rst_n:
            self._reg_array = 0
        elif self._clk_en:
            self._reg_array[0] = self._data_in
            for i in range(self.stages - 1):
                self._reg_array[i + 1] = self._reg_array[i]

    @always_comb
    def set_out(self):
        # Combinationally set this to the last register in the stage array.
        self._data_out = self._reg_array[self.stages - 1]

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_out_one_stage(self):
        if ~self._rst_n:
            self._reg_ = 0
        elif self._clk_en:
            self._reg_ = self._data_in


if __name__ == "__main__":
    stages = 3
    piperegdut = PipeReg(16, stages)
    verilog(piperegdut, filename="pipe_reg.sv")



import kratos as kts


class ConfigRegAttr(kts.Attribute):
    def __init__(self,
                 doc_string=""):
        super().__init__()
        self.value = "config_reg"
        self.documentation = doc_string

    def set_documentation(self, new_doc):
        self.documentation = new_doc

    def get_documentation(self):
        return self.documentation
